import os.path
import typing

import requests

from ..exceptions import StoreNotReachable, RepositoryInvalidConfig
from ..functions import modified_within
from ..store import Store
from ..commit import Commit


class JsonbinStore(Store):
    """Implementation using JSONBin for storage."""

    def __init__(self, managed_repository):
        super().__init__(managed_repository)
        self._url: str = self._managed_repository.config.get("store_url", "")
        self._headers: dict = self._managed_repository.config.get("store_headers", {})
        if not isinstance(self._headers, dict):
            raise RepositoryInvalidConfig()
        self._timeout: float = 5

    @property
    def _local_json_path(self) -> str:
        return os.path.join(
            self._managed_repository.working_dir, ".gitalong", "commits.json"
        )

    @property
    def commits(self) -> typing.List[Commit]:
        headers = {}
        for key, value in self._headers.items():
            headers[key] = os.path.expandvars(value)
        pull_threshold = self._managed_repository.config.get("pull_threshold", 60)
        serializable_commits = []
        if modified_within(self._local_json_path, pull_threshold):
            serializable_commits = self._read_local_json()
        response = requests.get(self._url, headers=headers, timeout=self._timeout)
        if response.status_code != 200:
            raise StoreNotReachable(response.status_code, response.text)
        serializable_commits = response.json()["record"]
        self._write_local_json(serializable_commits)
        return self._serializeables_to_commits(serializable_commits)

    @commits.setter
    def commits(self, commits: typing.List[Commit]):
        headers = {}
        for key, value in self._headers.items():
            headers[key] = os.path.expandvars(value)
        headers.update({"Content-Type": "application/json"})
        response = requests.put(
            self._url, headers=headers, json=commits, timeout=self._timeout
        )
        if response.status_code == 200:
            self._write_local_json(commits)
        else:
            raise StoreNotReachable()



# pylint: disable=missing-class-docstring

import os
import tempfile

from git.repo import Repo

from .cases import GitalongCase


class GitStoreTestCase(GitalongCase):

    __test__ = True

    def setUp(self):
        temp_dir = tempfile.mkdtemp()
        store_url = os.path.join(temp_dir, "store.git")
        Repo.init(path=store_url, bare=True)
        self._setup_repository(temp_dir, store_url)



class RepositoryNotSetup(Exception):
    """Error for when gitalong is not setup in the managed repository."""


class RepositoryInvalidConfig(Exception):
    """Error for when gitalong config is not as expected."""


class StoreNotReachable(Exception):
    """Error for when the store is not reachable."""


class CommandError(Exception):
    """Error for when running a command."""



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

import os
import shutil
import tempfile
import unittest
import logging

from gitalong.functions import (
    is_binary_file,
    get_filenames_from_move_string,
)

from .functions import save_image


class FunctionsTestCase(unittest.TestCase):
    """Sets up a temporary git repository for each test"""

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        try:
            shutil.rmtree(self.temp_dir)
        except PermissionError as error:
            logging.error(error)

    def test_is_binary_file(self):
        self.assertEqual(False, is_binary_file(__file__))
        image_path = os.path.join(self.temp_dir, "image.jpg")
        save_image(image_path)
        self.assertEqual(True, is_binary_file(image_path))

    def test_get_filenames_from_move_string(self):
        move_string = get_filenames_from_move_string("A/B/C.abc")
        self.assertEqual(("A/B/C.abc",), move_string)

        move_string = get_filenames_from_move_string("A/B/{C.abc => D.abc}")
        self.assertEqual(("A/B/C.abc", "A/B/D.abc"), move_string)

        move_string = get_filenames_from_move_string("A/B/{C..abc => C.abc}")
        self.assertEqual(("A/B/C..abc", "A/B/C.abc"), move_string)

        move_string = get_filenames_from_move_string("A/B/{C/D.abc => E/F.abc}")
        self.assertEqual(("A/B/C/D.abc", "A/B/E/F.abc"), move_string)



import os
import json
import asyncio

from gitalong import Repository
from gitalong.batch import get_files_last_commits


def test_local_files():
    """A test dedicated to local files. It expects a JSON file mapping path to
    commit spreads as shown below.

    {
        "C:\\file\\uncomitted": 1,
        "C:\\file\\in\\local\\active\\branch\\only": 2
    }
    """

    # Parse JSON file to get file paths and corresponding assertions.
    json_path = __file__.replace(".py", ".json")
    files = []
    if os.path.exists(json_path):
        with open(json_path, "r", encoding="utf-8") as f:
            files = json.loads(f.read())
    else:
        return

    filenames = files.keys()
    last_commits = asyncio.run(get_files_last_commits(filenames))

    for filename, last_commit in zip(filenames, last_commits):
        repository = Repository.from_filename(filename)
        spread = last_commit.commit_spread if repository else 0
        assert files[filename] == spread



import json
import os
import typing

import git
import git.exc
from git.repo import Repo

from ..exceptions import RepositoryInvalidConfig
from ..functions import pulled_within
from ..store import Store
from ..commit import Commit


class GitStore(Store):
    """Implementation using a Git repository as storage."""

    def __init__(self, managed_repository):
        super().__init__(managed_repository)
        self._store_repository = self._clone()

    @property
    def _local_json_path(self) -> str:
        return os.path.join(self._store_repository.working_dir, "commits.json")

    def _clone(self):
        """
        Returns:
            git.Repo: Clones the Gitalong repository if not done already.
        """
        try:
            return Repo(os.path.join(self._managed_repository.working_dir, ".gitalong"))
        except (git.exc.NoSuchPathError, git.exc.InvalidGitRepositoryError):
            try:
                remote = self._managed_repository.config["store_url"]
            except KeyError as error:
                config = self._managed_repository.config_path
                message = f'Could not find value for "store_url" in "{config}".'
                raise RepositoryInvalidConfig(message) from error
            return Repo.clone_from(
                remote,
                os.path.join(self._managed_repository.working_dir, ".gitalong"),
            )

    @property
    def commits(self) -> typing.List[Commit]:
        """
        Returns:
            typing.List[Commit]:
                A list of commits that haven't been pushed to remote. Also includes
                commits representing uncommitted changes.
        """
        store_repository = self._store_repository
        remote = store_repository.remote()
        pull_threshold = self._managed_repository.config.get("pull_threshold", 60)
        if not pulled_within(store_repository, pull_threshold) and remote.refs:
            # TODO: We could check that a pull is already happening.
            # This would avoid the try except and save time.
            try:
                remote.pull(
                    ff=True,
                    quiet=True,
                    rebase=True,
                    autostash=True,
                    verify=False,
                    summary=False,
                )
            except git.exc.GitCommandError:
                pass
        serializable_commits = self._read_local_json()
        if os.path.exists(self._local_json_path):
            with open(self._local_json_path, "r", encoding="utf-8") as _file:
                serializable_commits = json.loads(_file.read())
        return self._serializeables_to_commits(serializable_commits)

    @commits.setter
    def commits(self, commits: typing.List[Commit]):
        self._write_local_json(commits)
        self._store_repository.index.add(self._local_json_path)
        basename = os.path.basename(self._local_json_path)
        self._store_repository.index.commit(message=f"Update {basename}")
        self._store_repository.remote().push()



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=attribute-defined-outside-init

import tempfile
from unittest.mock import patch, MagicMock

from .cases import GitalongCase


class JsonbinStoreTestCase(GitalongCase):

    __test__ = True

    def setUp(self):
        temp_dir = tempfile.mkdtemp()
        url = "https://api.jsonbin.io/v3/b/<BIN_ID>/"

        self._stored_value = {"record": {}}

        self._get_patcher = patch("requests.get", self._get_patch)
        self._get_patcher.start()

        self._put_patcher = patch("requests.put", self._put_patch)
        self._put_patcher.start()

        self._store_headers = {"X-Master-Key": "<ACCESS_KEY>"}

        self._setup_repository(temp_dir, url, store_headers=self._store_headers)

    def _get_patch(
        self, url, headers=None, timeout=0
    ):  # pylint: disable=unused-argument
        self.assertEqual(url, self._store_url)
        self.assertDictEqual(headers or {}, self._store_headers)
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = self._stored_value
        return mock_response

    def _put_patch(
        self, url, headers=None, timeout=0, json=None
    ):  # pylint: disable=unused-argument
        self.assertEqual(url, self._store_url)
        store_headers = {"Content-Type": "application/json"}
        store_headers.update(self._store_headers)
        self.assertDictEqual(headers or {}, store_headers)
        self._stored_value = {"record": json or {}}
        mock_response = MagicMock()
        mock_response.status_code = 200
        return mock_response

    def tearDown(self):
        super().tearDown()
        self._get_patcher.stop()
        self._put_patcher.stop()



import asyncio

import click
import git
import git.exc

from .__info__ import __version__
from .enums import CommitSpread
from .repository import Repository
from .batch import get_files_last_commits


def get_status_string(filename: str, commit: dict, spread: int) -> str:
    """Generate a status string for a file and its commit."""
    prop = "+" if spread & CommitSpread.MINE_UNCOMMITTED else "-"
    prop += "+" if spread & CommitSpread.MINE_ACTIVE_BRANCH else "-"
    prop += "+" if spread & CommitSpread.MINE_OTHER_BRANCH else "-"
    prop += "+" if spread & CommitSpread.REMOTE_MATCHING_BRANCH else "-"
    prop += "+" if spread & CommitSpread.REMOTE_OTHER_BRANCH else "-"
    prop += "+" if spread & CommitSpread.THEIR_OTHER_BRANCH else "-"
    prop += "+" if spread & CommitSpread.THEIR_MATCHING_BRANCH else "-"
    prop += "+" if spread & CommitSpread.THEIR_UNCOMMITTED else "-"
    splits = [
        prop,
        filename,
        commit.get("sha", "-"),
        ",".join(commit.get("branches", {}).get("local", ["-"])) or "-",
        ",".join(commit.get("branches", {}).get("remote", ["-"])) or "-",
        commit.get("host", "-"),
        commit.get("author", commit.get("user", "-")) or "-",
    ]
    return " ".join(splits)


def validate_key_value(ctx, param, value):  # pylint: disable=unused-argument
    """Validate that the provided value is a valid key-value."""
    result = {}
    for item in value:
        key, val = item.split("=")
        result[key] = val
    return result


@click.command(help="Prints the requested configuration property value.")
def version():  # pylint: disable=missing-function-docstring
    click.echo(f"gitalong version {__version__}")


@click.command(help="Prints the requested configuration property value.")
@click.argument("prop")
@click.pass_context
def config(ctx, prop):  # pylint: disable=missing-function-docstring
    repository = Repository.from_filename(ctx.obj.get("REPOSITORY", ""))
    repository_config = repository.config if repository else {}
    prop = prop.replace("-", "_")
    if prop in repository_config:
        value = repository_config[prop]
        if isinstance(value, bool):
            value = str(value).lower()
        click.echo(value)


@click.command(
    help=(
        "Update tracked commits with the local changes of this clone. echo a list of "
        "files their permissions changed."
    )
)
@click.pass_context
def update(ctx):
    """Update tracked commits with local changes."""
    repository = Repository.from_filename(ctx.obj.get("REPOSITORY", ""))
    if not repository:
        return
    repository.update_tracked_commits()


@click.command(
    help=(
        "Prints missing commits in this local branch for each filename. "
        "Format: `<spread> <filename> <commit> <local-branches> "
        "<remote-branches> <host> <author>`"
    )
)
@click.argument("filename", nargs=-1)
@click.option(
    "-p",
    "--profile",
    is_flag=True,
    help=(
        "Will generate a profile file in the current working directory."
        "The file can be opened in a profiler like snakeviz."
    ),
)
@click.pass_context
def status(ctx, filename, profile=False):  # pylint: disable=missing-function-docstring
    if profile:
        import cProfile  # pylint: disable=import-outside-toplevel
        import pstats  # pylint: disable=import-outside-toplevel

        with cProfile.Profile() as pr:
            run_status(ctx, filename)
        results = pstats.Stats(pr)
        results.dump_stats("gitalong.prof")
        return
    run_status(ctx, filename)


def run_status(ctx, filename):  # pylint: disable=missing-function-docstring
    file_status = []
    commits = asyncio.run(get_files_last_commits(filename))
    for _filename, commit in zip(filename, commits):
        repository = Repository.from_filename(ctx.obj.get("REPOSITORY", _filename))
        absolute_filename = (
            repository.get_absolute_path(_filename) if repository else _filename
        )
        spread = commit.commit_spread if repository else 0
        file_status.append(get_status_string(absolute_filename, commit, spread))
    click.echo("\n".join(file_status), err=False)


@click.command(help="Setup Gitalong in a repository.")
@click.argument("store-url", required=True)
@click.option(
    "-sh",
    "--store-header",
    callback=validate_key_value,
    help="If using JSONBin.io as a store, the headers used to connect the endpoint.",
    required=False,
    multiple=True,
)
@click.option(
    "-mp",
    "--modify-permissions",
    is_flag=True,
    help=(
        "Whether or not Gitalong should affect file permissions of tracked files "
        "to prevent editing of files that are modified elsewhere. This is too "
        "expensive an option for repositories with many files and should be "
        "enabled."
    ),
)
@click.option(
    "-pt",
    "--pull-threshold",
    default=60,
    help=(
        "Time in seconds that need to pass before Gitalong pulls again. Defaults to 60"
        "seconds. This is for optimization sake as pull and fetch operations are "
        "expensive."
    ),
    required=False,
)
@click.option(
    "-tb",
    "--track-binaries",
    is_flag=True,
    help=(
        "Gitalong can track all auto-detected binary files "
        "to prevent conflicts on them. There is a performance cost to this feature so "
        "it's always better if you can specify the extensions you care about tracking "
        "using --tracked-extensions."
    ),
    required=False,
)
@click.option(
    "-tu",
    "--track-uncommitted",
    is_flag=True,
    help=(
        "Track uncommitted changes. Better for collaboration but requires pushing"
        "tracked commits after each file system operation."
    ),
)
@click.option(
    "-te",
    "--tracked-extensions",
    default="",
    help=("A comma-separated list of extensions to track to prevent conflicts."),
)
@click.option(
    "-ug",
    "--update-gitignore",
    is_flag=True,
    help=(".gitignore should be modified in the repository to ignore Gitalong files."),
)
@click.option(
    "-uh",
    "--update-hooks",
    is_flag=True,
    help="Hooks should be updated with Gitalong logic.",
)
@click.pass_context
def setup(  # pylint: disable=too-many-positional-arguments
    ctx,
    store_url,
    store_header,
    modify_permissions,
    pull_threshold,
    track_binaries,
    track_uncommitted,
    tracked_extensions,
    update_gitignore,
    update_hooks,
):
    """Setup Gitalong in a repository."""
    Repository.setup(
        store_url=store_url,
        store_headers=store_header,
        managed_repository=ctx.obj.get("REPOSITORY", ""),
        modify_permissions=modify_permissions,
        pull_threshold=pull_threshold,
        track_binaries=track_binaries,
        track_uncommitted=track_uncommitted,
        tracked_extensions=tracked_extensions.split(","),
        update_gitignore=update_gitignore,
        update_hooks=update_hooks,
    )


class Group(click.Group):  # pylint: disable=missing-class-docstring
    def format_help(self, ctx, formatter):
        return click.Group.format_help(self, ctx, formatter)


@click.group(cls=Group)
@click.version_option(
    prog_name="gitalong",
    version=__version__,
    message="%(prog)s version %(version)s",
)
@click.option(
    "-C",
    "--repository",
    default="",
    help=(
        "The repository to apply operations to. "
        "Defaults to current working directory."
    ),
    required=False,
)
@click.option(
    "-gb",
    "--git-binary",
    default="",
    help="Path to the git binary to use. Defaults to the one available in PATH.",
    required=False,
)
@click.pass_context
def cli(ctx, repository, git_binary):  # pylint: disable=missing-function-docstring
    ctx.ensure_object(dict)
    if repository:
        ctx.obj["REPOSITORY"] = repository
    if git_binary:
        git.refresh(git_binary)


cli.add_command(config)
cli.add_command(update)
cli.add_command(setup)
cli.add_command(status)
cli.add_command(version)


def main():
    """This main function will be registered as the console script when installing the
    package.
    """
    cli(obj={})  # pylint: disable=unexpected-keyword-arg,no-value-for-parameter


if __name__ == "__main__":
    main()



import socket
import getpass

from typing import List

import dictdiffer
import git
import git.exc

from git import Repo


from .functions import get_real_path
from .enums import CommitSpread


class Commit(dict):
    """A class to represent a commit object in Git."""

    def __init__(self, repository, *args, **kwargs):
        self._repository = repository
        self._git_commit = None
        super().__init__(*args, **kwargs)

    @property
    def repository(self):
        """Return the repository object that this commit belongs to."""
        return self._repository

    def update_with_git_commit(self, git_commit: git.Commit):
        """Update the commit using info from a GitPython commit object.

        Args:
            git_commit (git.Commit): _description_
        """
        self.update(
            {
                "remote": git_commit.repo.remote().url,
                "date": str(git_commit.committed_datetime),
                "author": git_commit.author.name,
                "clone": git_commit.repo.working_dir,
            }
        )

    def update_with_sha(self, sha):
        """
        Returns:
            dict: A dict of contextual values that we attached to tracked commits.
        """
        self["sha"] = sha

        if not self._repository:
            return
        try:
            git_commit = Repo(self._repository.working_dir).commit(sha)
        except git.exc.BadName:
            git_commit = None
        if git_commit:
            self.update_with_git_commit(git_commit)

    def update_context(self):
        """
        Returns:
            dict: A dict of contextual values that we attached to tracked commits.
        """
        self.update(self.context_dict)

    @property
    def context_dict(self) -> dict:
        """
        Returns:
            dict: A dict of contextual values that we attached to tracked commits.
        """
        context_dict = {"host": socket.gethostname(), "user": getpass.getuser()}
        if self._repository:
            context_dict["clone"] = get_real_path(self._repository.working_dir)
        return context_dict

    def is_issued_commit(self) -> bool:
        """
        Args:
            commit (dict): The commit dictionary to check for.

        Returns:
            bool: Whether the commit was issued by the current context.
        """
        context_dict = self.context_dict
        diff_keys = set()
        for diff in dictdiffer.diff(context_dict, self):
            if diff[0] == "change":
                diff_keys.add(diff[1])
            elif diff[0] in ("add", "remove"):
                diff_keys = diff_keys.union([key[0] for key in diff[2]])
        intersection = set(context_dict.keys()).intersection(diff_keys)
        return not intersection

    def is_issued_uncommitted_changes_commit(self) -> bool:
        """
        Args:`
            commit (dict): Description

        Returns:
            bool:
                Whether the commit represents uncommitted changes and is issued by the
                current context.
        """
        if not self.is_uncommitted_changes_commit():
            return False
        return self.is_issued_commit()

    def is_uncommitted_changes_commit(self) -> bool:
        """
        Args:
            commit (dict): The commit dictionary.

        Returns:
            bool: Whether the commit dictionary represents uncommitted changes.
        """
        return "user" in self.keys()

    @property
    def parents(self) -> List[str]:
        """
        Returns:
            Commit: The parent commit of the current commit.
        """
        parents = []
        if self._repository:
            self._git_commit = self._git_commit or Repo(
                self._repository.working_dir
            ).commit(self.get("sha"))
            if self._git_commit:
                for parent_git_commit in self._git_commit.parents:
                    parent = Commit(self._repository)
                    parent.update_with_git_commit(parent_git_commit)
                    parents.append(parent)
        return parents

    @property
    def commit_spread(self) -> int:
        """
        Args:
            commit (int): The commit to check for.

        Returns:
            dict:
                A dictionary of commit spread information containing all
                information about where this commit lives across branches and clones.
        """
        commit_spread = 0
        active_branch = self._repository.active_branch_name
        if self.get("user", ""):
            is_issued = self.is_issued_commit()
            if "sha" in self:
                if active_branch in self.get("branches", {}).get("local", []):
                    commit_spread |= (
                        CommitSpread.MINE_ACTIVE_BRANCH
                        if is_issued
                        else CommitSpread.THEIR_MATCHING_BRANCH
                    )
                else:
                    commit_spread |= (
                        CommitSpread.MINE_OTHER_BRANCH
                        if is_issued
                        else CommitSpread.THEIR_OTHER_BRANCH
                    )
            else:
                commit_spread |= (
                    CommitSpread.MINE_UNCOMMITTED
                    if is_issued
                    else CommitSpread.THEIR_UNCOMMITTED
                )
        else:
            remote_branches = self.get("branches", {}).get("remote", [])
            if active_branch in remote_branches:
                commit_spread |= CommitSpread.REMOTE_MATCHING_BRANCH
            if active_branch in self.get("branches", {}).get("local", []):
                commit_spread |= CommitSpread.MINE_ACTIVE_BRANCH
            if active_branch in remote_branches:
                remote_branches.remove(active_branch)
            if remote_branches:
                commit_spread |= CommitSpread.REMOTE_OTHER_BRANCH
        return commit_spread



import os
import stat
import asyncio

# import datetime

# from turtle import write
from typing import List, Coroutine

# from unittest import result

import git
import git.exc

# from more_itertools import last

from .enums import CommitSpread
from .repository import Repository
from .commit import Commit

# from .functions import is_binary_file
from .exceptions import CommandError


async def _run_command(args: List[str], safe: bool = False) -> str:
    """
    Args:
        args (List[str]): The command to run.
        safe (bool): If True, suppress exceptions and return an empty string on failure.

    Raises:
        Exception: When the command fails and safe is False.

    Returns:
        str: The stdout of the command.
    """
    process = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.DEVNULL,
    )
    stdout, stderr = await process.communicate()
    if process.returncode != 0:
        if safe:
            return ""
        error = stderr.decode().strip()
        raise CommandError(f"Command {' '.join(args)} failed with error: {error}")
    return stdout.decode().strip()


async def get_files_last_commits(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    filenames: List[str], prune: bool = True
) -> List[Commit]:
    """Get the last commit for a list of files.

    Args:
        filenames (List[str]): A list of absolute filenames to get the last commit for.
        prune (bool): Prune branches if a fetch is necessary.

    Returns:
        List[Commit]: A list of last commits for the files.
    """
    pruned_repositories: List[str] = []
    last_commits: List[Commit] = []
    for filename in filenames:
        repository = Repository.from_filename(os.path.dirname(filename))
        last_commit = Commit(repository)

        if not repository:
            last_commits.append(last_commit)
            continue

        if not repository.is_file_tracked(filename):
            last_commits.append(last_commit)
            continue

        # We are checking the tracked commit first as they represented local changes.
        # They are in nature always more recent. If we find a relevant commit here we
        # can skip looking elsewhere.
        tracked_commits = repository.store.commits
        relevant_tracked_commits = []
        filename = repository.get_relative_path(filename)
        remote = repository.remote_url
        track_uncommitted = repository.config.get("track_uncommitted", False)
        for tracked_commit in tracked_commits:
            if (
                # We ignore uncommitted tracked commits if configuration says so.
                (not track_uncommitted and "sha" not in tracked_commit)
                # We ignore commits from other remotes.
                or tracked_commit.get("remote") != remote
            ):
                continue
            for change in tracked_commit.get("changes", []):
                if os.path.normpath(change) == os.path.normpath(filename):
                    relevant_tracked_commits.append(tracked_commit)
                    continue
        if relevant_tracked_commits:
            relevant_tracked_commits.sort(key=lambda commit: commit.get("date"))
            last_commit = relevant_tracked_commits[-1]

            # Because there is no post-push hook a local commit that got pushed could
            # have never been removed from our tracked commits. To cover for this case
            # we are checking if this commit is on remote and modify it, so it's
            # conform to a remote commit.
            branches_list = await get_commits_branches([last_commit], remote=True)
            if "sha" in last_commit and branches_list[0]:
                tracked_commits.remove(last_commit)
                repository.store.commits = tracked_commits
                for key in repository.context_dict:
                    if key in last_commit:
                        del last_commit[key]

        if not last_commit:
            pull_threshold = repository.config.get("pull_threshold", 60)
            if not repository.pulled_within(pull_threshold):
                try:
                    # We fetch only once since it's costly.
                    working_dir = repository.working_dir
                    if working_dir not in pruned_repositories:
                        repository.remote.fetch(prune=prune)
                        pruned_repositories.append(working_dir)
                except git.exc.GitCommandError:
                    pass

            args = [
                "--all",
                "--remotes",
                '--pretty=format:"%H"',
                # TODO: It is said that the chronological order of commits in the commit
                # history does not necessarily reflect the order of their commit dates.
                "--date-order",
                "--",
                filename,
            ]
            output = repository.git.log(*args)
            file_commits = output.replace('"', "").split("\n") if output else []
            sha = file_commits[0] if file_commits else ""
            last_commit = Commit(repository)
            last_commit.update_with_sha(sha)

        last_commits.append(last_commit)

    changes = await get_commits_changes(last_commits)
    local_branches_list = await get_commits_branches(last_commits)
    remote_branches_list = await get_commits_branches(last_commits, remote=True)

    for last_commit, changes, local_branches, remote_branches in zip(
        last_commits, changes, local_branches_list, remote_branches_list
    ):
        if changes:
            last_commit["changes"] = changes
        if local_branches:
            branches = last_commit.setdefault("branches", {})
            branches["local"] = local_branches
        if remote_branches:
            branches = last_commit.setdefault("branches", {})
            branches["remote"] = remote_branches

    return last_commits


async def get_commits_branches(
    commits: List[Commit], remote: bool = False
) -> List[str]:
    """
    Args:
        commits (List[Commit]): The commits to check for branches.
        remote (bool, optional): Whether we should return local or remote branches.

    Returns:
        List[str]: A list of branch names that this commit is living on.
    """
    branches_list = []
    tasks = []
    for commit in commits:
        if "sha" not in commit:
            # We are going to run a fake task to keep the order of the results.
            task = asyncio.sleep(0)
        else:
            args = ["git", "-C", commit.repository.working_dir, "branch"]
            if remote:
                args += ["--remote"]
            args += ["--contains", commit.get("sha", "")]
            task = _run_command(args, safe=True)
        tasks.append(task)
    results = await asyncio.gather(*tasks)
    for commit, result in zip(commits, results):
        if "sha" not in commit:
            branches_list.append([])
        else:
            branches = result.replace("*", "")
            branches = branches.replace(" ", "")
            branches = branches.split("\n") if branches else []
            branch_names = set()
            for branch in branches:
                branch_names.add(branch.split("/")[-1])
            branches_list.append(list(branch_names))
    return branches_list


async def get_commits_changes(commits: List[Commit]) -> List[str]:
    """
    Args:
        commits (List[Commit]): The commits to get the changes from.

    Returns:
        List[str]: A list of filenames that have changed in the commit.
    """
    changes_list = []
    tasks: List[Coroutine] = []
    for commit in commits:
        if "changes" in commit:
            changes_list.append(commit["changes"])
            continue

        repository = commit.repository
        if not repository:
            changes_list.append([])
            continue

        sha = commit.get("sha", "")
        if not sha:
            changes_list.append([])
            continue

        # If changes have not been collected we do that.
        if not commit.parents:
            # For the first commit, use git show.
            args = [
                "git",
                "-C",
                repository.working_dir,
                "show",
                "--pretty=format:",
                "--name-only",
                sha,
            ]
        else:
            # For subsequent commits, using git diff-tree.
            args = [
                "git",
                "-C",
                repository.working_dir,
                "diff-tree",
                "--no-commit-id",
                "--name-only",
                "-r",
                sha,
            ]
        tasks.append(_run_command(args))
        changes_list.append((None))

    results = await asyncio.gather(*tasks)
    for result in results:
        changes = result.split("\n")
        changes = [change for change in changes if change]
        index = changes_list.index(None)
        changes_list[index] = changes
    return changes_list


async def update_files_permissions(filenames: List[str]):
    """Updates the permissions of a file weather or not they can be changed.

    Args:
        filename (str): The relative or absolute filename to update permissions for.
    """
    tasks = []
    last_commits = await get_files_last_commits(filenames)
    write_permissions = []
    for filename, last_commit in zip(filenames, last_commits):
        repository = Repository.from_filename(os.path.dirname(filename))
        spread = last_commit.commit_spread if repository else 0
        if not spread:
            continue
        is_uncommitted = (
            spread & CommitSpread.MINE_UNCOMMITTED == CommitSpread.MINE_UNCOMMITTED
        )
        is_local = (
            spread & CommitSpread.MINE_ACTIVE_BRANCH == CommitSpread.MINE_ACTIVE_BRANCH
        )
        is_current = (
            spread
            & (CommitSpread.MINE_ACTIVE_BRANCH | CommitSpread.REMOTE_MATCHING_BRANCH)
            == CommitSpread.MINE_ACTIVE_BRANCH | CommitSpread.REMOTE_MATCHING_BRANCH
        )
        write_permission = is_uncommitted or is_local or is_current
        write_permissions.append(write_permission)
        tasks.append(_set_write_permission(filename, write_permission))
    await asyncio.gather(*tasks)


async def _set_write_permission(
    filename: str, write_permission: bool, safe: bool = False
) -> bool:
    """
    Set the write permission of a file asynchronously.

    Args:
        filename (str): The path to the file.
        write_permission (bool): Whether to make the file writable.
        safe (bool): Won't raise an exceptions if the file doesn't exist or is
        not accessible.

    Returns:
        bool: Whether the file ends in the desired state.
    """
    try:
        current_permissions = os.stat(filename).st_mode
        if write_permission:
            os.chmod(filename, current_permissions | stat.S_IWRITE)
        else:
            os.chmod(filename, current_permissions & ~stat.S_IWRITE)
    except FileNotFoundError:
        if safe:
            # If the file doesn't exist we can't set it's permissions.
            return False
        raise
    except PermissionError:
        if safe:
            if write_permission:
                return False
            return True
        raise
    return True



from .__info__ import __version__, __copyright__, __email__, __author__
from .enums import CommitSpread
from .exceptions import RepositoryNotSetup
from .repository import Repository
from .commit import Commit



import configparser
import datetime
import getpass
import json
import logging
import os
import shutil
import socket
import asyncio

from typing import Optional, List

import git
import git.exc

from git.repo import Repo

from .store import Store
from .stores.git_store import GitStore
from .stores.jsonbin_store import JsonbinStore
from .exceptions import RepositoryNotSetup, RepositoryInvalidConfig
from .functions import (
    get_real_path,
    is_binary_file,
    pulled_within,
)


class Repository:  # pylint: disable=too-many-public-methods
    """Aggregates all the Gitalong actions that can happen on a Git repository.

    Raises:
        git.exc.InvalidGitRepositoryError: If the path is not in or the Git clone.
    """

    _instances = {}
    _config_basename = ".gitalong.json"

    def __new__(
        cls,
        repository: str = "",
        use_cached_instances=False,
    ):
        managed_repo = Repo(repository, search_parent_directories=True)
        working_dir = managed_repo.working_dir
        if use_cached_instances:
            return cls._instances.setdefault(working_dir, super().__new__(cls))
        return super().__new__(cls)

    def __init__(
        self,
        repository: str = "",
        use_cached_instances=False,  # pylint: disable=unused-argument
    ):
        """
        Args:
            repository (str):
                The managed repository exact absolute path.
            use_cached_instances (bool):
                If true, the class will return "singleton" cached per clone.

        Raises:
            RepositoryInvalidConfig: If the store URL is not valid.
            RepositoryNotSetup: If Gitalong is not installed on the repository.
        """
        self._config = None
        self._submodules = None

        self._managed_repository = Repo(repository, search_parent_directories=True)
        self._remote = self._managed_repository.remote()

        store_url = self.config.get("store_url", "")
        if store_url.startswith("https://api.jsonbin.io"):
            self._store = JsonbinStore(self)
        elif store_url.endswith(".git"):
            self._store = GitStore(self)
        else:
            raise RepositoryInvalidConfig("Invalid store URL in configuration.")

        if self.config.get(
            "modify_permissions", False
        ) and self._managed_repository.config_reader().get_value(
            "core", "fileMode", True
        ):
            config_writer = self._managed_repository.config_writer()
            config_writer.set_value("core", "fileMode", "false")
            config_writer.release()

    @classmethod
    def setup(  # pylint: disable=too-many-positional-arguments
        cls,
        store_url: str,
        store_headers: Optional[dict] = None,
        managed_repository: str = "",
        modify_permissions=False,
        pull_threshold: float = 60.0,
        track_binaries: bool = False,
        track_uncommitted: bool = False,
        tracked_extensions: Optional[List[str]] = None,
        update_gitignore: bool = False,
        update_hooks: bool = False,
    ):
        """Setup Gitalong in a repository.

        Args:
            store_url (str):
                The URL or path to the repository that Gitalong will use to store local
                changes.
            store_headers (str):
                The headers to connect to the API end point.
            managed_repository (str, optional):
                The repository in which we install Gitalong. Defaults to current
                working directory. Current working directory if not passed.
            modify_permissions (bool, optional):
                Whether Gitalong should manage permissions of binary files.
            track_binaries (bool, optional):
                Track all binary files by automatically detecting them.
            track_uncommitted (bool, optional):
                Track uncommitted changes. Better for collaboration but requires to push
                tracked commits after each file system operation.
            tracked_extensions (List[str], optional):
                List of extensions to track.
            pull_threshold (list, optional):
                Time in seconds that need to pass before Gitalong pulls again. Defaults
                to 10 seconds. This is for optimization's sake as pull and fetch
                operation are expensive. Defaults to 60 seconds.
            update_gitignore (bool, optional):
                Whether .gitignore should be modified in the managed repository to
                ignore Gitalong files.
            update_hooks (bool, optional):
                Whether hooks should be updated with Gitalong logic.

        Returns:
            Gitalong:
                The Gitalong management class corresponding to the repository in
                which we just installed.
        """
        tracked_extensions = tracked_extensions or []
        managed_repo = Repo(managed_repository, search_parent_directories=True)
        config_path = os.path.join(managed_repo.working_dir, cls._config_basename)
        config = {
            "store_url": store_url,
            "store_headers": store_headers or {},
            "modify_permissions": modify_permissions,
            "track_binaries": track_binaries,
            "tracked_extensions": tracked_extensions,
            "pull_threshold": pull_threshold,
            "track_uncommitted": track_uncommitted,
        }
        cls._write_config_file(config, config_path)
        gitalong = cls(repository=str(managed_repo.working_dir))
        if modify_permissions:
            managed_repo.config_writer().set_value("core", "fileMode", "true").release()
        if update_gitignore:
            gitalong.update_gitignore()
        if update_hooks:
            gitalong.install_hooks()
        return gitalong

    @classmethod
    def from_filename(cls, filename: str) -> Optional["Repository"]:
        """
        Args:
            filename (str):
                Existing absolute path to a file or folder in the managed repository.
                That includes the managed repository itself.

        Returns:
            Optional[Repository]: The repository or None.
        """
        try:
            return cls(repository=filename, use_cached_instances=True)
        except (
            git.exc.InvalidGitRepositoryError,
            RepositoryNotSetup,
            git.exc.NoSuchPathError,
        ):
            return None

    @staticmethod
    def _write_config_file(config: dict, path: str):
        with open(path, "w", encoding="utf8") as config_file:
            json.dump(config, config_file, indent=4, sort_keys=True)

    def update_gitignore(self):
        """Update the .gitignore of the managed repository with Gitalong directives."""
        # TODO: Improve by considering what is already ignored.
        gitignore_path = os.path.join(self.working_dir, ".gitignore")
        content = ""
        if os.path.exists(gitignore_path):
            with open(gitignore_path, encoding="utf8") as gitignore:
                content = gitignore.read()
        with open(gitignore_path, "w", encoding="utf8") as gitignore:
            with open(
                os.path.join(os.path.dirname(__file__), "resources", "gitignore"),
                encoding="utf8",
            ) as patch:
                patch_content = patch.read()
            if patch_content not in content:
                gitignore.write(content + patch_content)

    @property
    def config_path(self) -> str:
        """
        Returns:
            str: The path to the configuration file.
        """
        return os.path.join(self.working_dir, self._config_basename)

    @property
    def store(self) -> Store:
        """
        Returns:
            Store: The store that Gitalong uses to keep track of local changes.
        """
        return self._store

    @property
    def config(self) -> dict:
        """
        Returns:
            dict: The content of `.gitalong.json` as a dictionary.
        """
        if self._config is None:
            try:
                with open(self.config_path, encoding="utf8") as _config_file:
                    self._config = json.loads(_config_file.read())
            except FileNotFoundError as error:
                raise RepositoryNotSetup(
                    "Gitalong is not installed on this repository."
                ) from error
        return self._config

    @property
    def remote(self) -> git.Remote:
        """
        Returns:
            git.Remote: The remote repository of the managed repository.
        """
        return self._remote

    @property
    def remote_url(self) -> str:
        """
        Returns:
            str: The URL of the remote repository.
        """
        return self._remote.url

    @property
    def hooks_path(self) -> str:
        """
        Returns:
            str: The hook path of the managed repository.
        """
        try:
            basename = self._managed_repository.config_reader().get_value(
                "core", "hooksPath"
            )
        except configparser.NoOptionError:
            basename = os.path.join(".git", "hooks")
        path = os.path.join(
            self.working_dir,
            basename,  # pyright: ignore[reportCallIssue, reportArgumentType]
        )
        return os.path.normpath(path)

    def install_hooks(self):
        """Installs Gitalong hooks in managed repository.

        TODO: Implement non-destructive version of these hooks. Currently we don't have
        any consideration for pre-existing content.
        """
        hooks = os.path.join(os.path.dirname(__file__), "resources", "hooks")
        destination_dir = self.hooks_path
        for dirname, _, basenames in os.walk(hooks):
            for basename in basenames:
                filename = os.path.join(dirname, basename)
                destination = os.path.join(destination_dir, basename)
                msg = f"Copying hook from {filename} to {destination}"
                logging.info(msg)
                shutil.copyfile(filename, destination)

    def get_relative_path(self, filename: str) -> str:
        """
        Args:
            filename (str): The absolute path.

        Returns:
            str: The path relative to the managed repository.
        """
        if os.path.exists(filename):
            filename = os.path.relpath(filename, self.working_dir)
        return filename

    def get_absolute_path(self, filename: str) -> str:
        """
        Args:
            filename (str): The path relative to the managed repository.

        Returns:
            str: The absolute file system path.
        """
        if os.path.exists(filename):
            return filename
        return os.path.join(self.working_dir, filename)

    @property
    def uncommitted_changes_commit(self) -> dict:
        """
        Returns:
            dict: Returns a commit dictionary representing uncommitted changes.
        """
        uncommitted_changes = self._uncommitted_changes
        if not uncommitted_changes:
            return {}

        commit = {
            "remote": self._remote.url,
            "changes": self._uncommitted_changes,
            "date": str(datetime.datetime.now()),
        }
        commit.update(self.context_dict)
        return commit

    def _accumulate_local_only_commits(
        self, start: git.Commit, local_commits: List[dict]
    ):
        """Accumulates a list of local only commit starting from the provided commit.

        Args:
            local_commits (list): The accumulated local commits.
            start (git.objects.Commit):
                The commit that we start peeling from last commit.
        """
        from .commit import Commit  # pylint: disable=import-outside-toplevel

        if self._managed_repository.git.branch("--remotes", "--contains", start.hexsha):
            return

        commit = Commit(self)
        commit.update_with_sha(start.hexsha)
        commit.update_context()

        changes = asyncio.run(self.batch.get_commits_changes([commit]))
        commit["changes"] = changes[0]

        branches_list = asyncio.run(self.batch.get_commits_branches([commit]))
        branches = branches_list[0] if branches_list else []
        commit["branches"] = {"local": branches}

        if commit not in local_commits:
            local_commits.append(commit)
        for parent in start.parents:
            self._accumulate_local_only_commits(parent, local_commits)

    @property
    def context_dict(self) -> dict:
        """
        Returns:
            dict: A dict of contextual values that we attached to tracked commits.
        """
        return {
            "host": socket.gethostname(),
            "user": getpass.getuser(),
            "clone": get_real_path(self.working_dir),
        }

    def get_local_only_commits(self) -> list:
        """
        Returns:
            list:
                Commits that are not on remote branches. Includes a commit that
                represents uncommitted changes.
        """
        local_commits = []
        for branch in self._managed_repository.branches:
            self._accumulate_local_only_commits(branch.commit, local_commits)
        if self.config.get("track_uncommitted"):
            uncommitted_changes_commit = self.uncommitted_changes_commit
            if uncommitted_changes_commit:
                local_commits.insert(0, uncommitted_changes_commit)
        local_commits.sort(key=lambda commit: commit.get("date"), reverse=True)
        return local_commits

    @property
    def _uncommitted_changes(self) -> list:
        """
        Returns:
            list: A list of unique relative filenames that feature uncommitted changes.
        """
        git_cmd = self._managed_repository.git
        output = git_cmd.ls_files("--exclude-standard", "--others")
        untracked_changes = output.split("\n") if output else []
        output = git_cmd.diff("--name-only")
        changes = output.split("\n") if output else []
        output = git_cmd.diff("--staged", "--name-only")
        staged_changes = output.split("\n") if output else []
        # A file can be in both in untracked and staged changes.
        return list(set(untracked_changes + changes + staged_changes))

    def _is_ignored(self, filename: str) -> bool:
        """
        Args:
            filename (str): The filename to check for.

        Returns:
            bool: Whether a file is ignored by the managed repository .gitignore file.
        """
        filename = self.get_relative_path(filename)
        try:
            self._managed_repository.git.check_ignore(filename)
            return True
        except git.exc.GitCommandError:
            return False

    @property
    def files(self) -> list:
        """
        Returns:
            list:
                The relative filenames that are tracked by the managed repository. Not
                to be confused with the files tracked by Gitalong.
        """
        git_cmd = self._managed_repository.git
        try:
            # TODO: HEAD might not be safe. The user could checkout an earlier commit.
            filenames = git_cmd.ls_tree(full_tree=True, name_only=True, r="HEAD")
            if not filenames:
                return []
            return filenames.split("\n")
        except git.exc.GitCommandError:
            return []

    @property
    def _locally_changed_files(self) -> list:
        """
        Returns:
            list:
                The relative filenames that have been changed by local commits or
                uncommitted changes.
        """
        local_changes = set()
        for commit in self.get_local_only_commits():
            local_changes = local_changes.union(commit.get("changes", []))
        return list(local_changes)

    def is_file_tracked(self, filename: str) -> bool:
        """
        Args:
            filename (str): The absolute or relative file or folder path to check for.

        Returns:
            bool: Whether the file is tracked by Gitalong.
        """
        if self._is_ignored(filename):
            return False
        tracked_extensions = self.config.get("tracked_extensions", [])
        if os.path.splitext(filename)[-1] in tracked_extensions:
            return True
        # The binary check is expensive, so doing it last.
        # TODO: If we check for a file that does not exist locally if will return False.
        # That file technically could be a tracked binary modifed somewhere else.
        return self.config.get("track_binaries", False) and is_binary_file(
            self.get_absolute_path(filename), safe=True
        )

    @property
    def working_dir(self) -> str:
        """
        Returns:
            str: The working directory of the managed repository.
        """
        return str(self._managed_repository.working_dir)

    def update_tracked_commits(self):
        """Pulls the tracked commits from the store and updates them."""
        self._store.commits = self._get_updated_tracked_commits()
        if self.config.get("modify_permissions"):
            absolute_filenames = []
            for filename in self.files:
                absolute_filenames.append(self.get_absolute_path(filename))
            asyncio.run(self.batch.update_files_permissions(absolute_filenames))

    def _get_updated_tracked_commits(self) -> list:
        """
        Returns:
            list:
                Local commits for all clones with local commits and uncommitted changes
                from this clone.
        """
        tracked_commits = []
        for commit in self._store.commits:
            remote = self._remote.url
            is_other_remote = commit.get("remote") != remote
            if is_other_remote or not commit.is_issued_commit():
                tracked_commits.append(commit)
                continue

        for commit in self.get_local_only_commits():
            tracked_commits.append(commit)
        return tracked_commits

    def pulled_within(self, seconds: float) -> bool:
        """
        Args:
            seconds (float): Time in seconds since last push.

        Returns:
            bool: Whether the repository pulled within the time provided.
        """
        return pulled_within(self._managed_repository, seconds)

    def log(self, message: str):
        """Logs a message to the managed repository.

        Args:
            message (str): The message to log.
        """
        self._managed_repository.git.log(message)

    @property
    def git(self) -> git.Git:
        """
        Returns:
            git.cmd.Git: The Git command line interface for the managed repository.
        """
        return self._managed_repository.git

    @property
    def batch(self):
        """
        Returns:
            Batch: The batch object for the managed repository.
        """
        from . import batch  # pylint: disable=import-outside-toplevel

        return batch

    @property
    def active_branch_name(self) -> str:
        """
        Returns:
            str: The name of the active branch.
        """
        return self._managed_repository.active_branch.name



import json
import os
import typing

from abc import ABC, abstractmethod

from .commit import Commit


class Store(ABC):
    """Abstract class for storing commits information."""

    def __init__(self, managed_repository):
        super().__init__()
        self._managed_repository = managed_repository

    def _read_local_json(self) -> typing.List[Commit]:
        if os.path.exists(self._local_json_path):
            with open(self._local_json_path, "r", encoding="utf-8") as _file:
                return json.loads(_file.read())
        return []

    def _write_local_json(self, commits: typing.List[Commit]):
        cache_dirname = os.path.dirname(self._local_json_path)
        if not os.path.exists(cache_dirname):
            os.makedirs(cache_dirname)
        with open(self._local_json_path, "w", encoding="utf-8") as _file:
            json.dump(commits, _file, indent=4, sort_keys=True)

    @property
    @abstractmethod
    def _local_json_path(self) -> str:
        """
        Returns:
            str: The path to the JSON file that tracks the local commits.
        """
        return ""

    @property
    @abstractmethod
    def commits(self) -> typing.List[Commit]:
        """
        Returns:
            typing.List[Commit]: The stored commits.
        """
        return []

    @commits.setter
    @abstractmethod
    def commits(self, commits: typing.List[Commit]):
        """Update the stored commits.

        Args: commits (list, optional): The commits to update the store with.
        """
        pass  # pylint: disable=unnecessary-pass

    def _serializeables_to_commits(self, serializable_commits):
        """Turn serializable commits into Commit objects."""
        commits = []
        for serializable_commit in serializable_commits:
            commit = Commit(self._managed_repository)
            commit.update(serializable_commit)
            commits.append(commit)
        return commits



import os
import pathlib
import re
import stat
import time

from git.repo import Repo


MOVE_STRING_REGEX = re.compile("{(.*)}")


def is_binary_file(filename: str, safe: bool = False) -> bool:
    """
    Args:
        filename (str): The path to the file to analyze.

    Returns:
        bool: Whether the file is a binary.
    """
    try:
        with open(filename, "rb") as fle:
            return is_binary_string(
                fle.read(1024)  # pyright: ignore[reportArgumentType]
            )
    except (IsADirectoryError, FileNotFoundError):
        if safe:
            return False
        raise


def is_binary_string(string: str) -> bool:
    """
    Args:
        string (str): A string to analyze.

    Returns:
        bool: Whether the string is a binary string.
    """
    textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})
    return bool(string.translate(None, textchars))  # pyright: ignore[reportCallIssue]


def is_writeable(filename: str) -> bool:
    """
    Args:
        filename (str): The absolute filename of the file to check.

    Returns:
        bool: Whether the file is read only.
    """
    stat_ = os.stat(filename)
    return bool(stat_.st_mode & stat.S_IWUSR)


def get_real_path(filename: str) -> str:
    """
    Args:
        filename (str): The filemame for which we want the real path.

    Returns:
        str: Real path in case this path goes through Windows subst.
    """
    # On Windows, this private function is available and will return the real path
    # for a subst location.
    if hasattr(os.path, "_getfinalpathname"):
        filename = os.path._getfinalpathname(  # pylint: disable=protected-access,line-too-long # pyright: ignore[reportAttributeAccessIssue]
            filename
        )
        filename = str(pathlib.Path(filename).resolve())
    return filename


def modified_within(filename: str, seconds: float) -> bool:
    """
    Args:
        filename (str): The file to check for.
        seconds (float): Time in seconds since last push.

    Returns:
        TYPE: Whether the file was modified within the time provided.
    """
    if not os.path.exists(filename):
        return False
    modified_time = os.path.getmtime(filename)
    current_time = time.time()
    return current_time - modified_time < seconds


def pulled_within(repository: Repo, seconds: float) -> bool:
    """
    Args:
        repository (Repo): The repository to check for.
        seconds (float): Time in seconds since last push.

    Returns:
        TYPE: Whether the repository pulled within the time provided.
    """
    fetch_head = os.path.join(repository.git_dir, "FETCH_HEAD")
    if not os.path.exists(fetch_head):
        return False
    since_last = time.time() - os.path.getmtime(fetch_head)
    return seconds > since_last


def get_filenames_from_move_string(move_string: str) -> tuple:
    """
    Args:
        move_string (str): The move string returned by git status.

    Returns:
        tuple: A tuple with the old and new filename of the moved file.
    """
    arrow = " => "
    if arrow not in move_string:
        return (move_string,)
    lefts = []
    rights = []
    match = MOVE_STRING_REGEX.search(move_string)
    if match:
        for group in match.groups():
            move_string = move_string.replace(group, "")
            splits = group.split(arrow)
            lefts.append(splits[0])
            rights.append(splits[-1])
    pair = {move_string.format(*lefts), move_string.format(*rights)}
    return tuple(sorted(pair))
